@@ -363,6 +363,7 @@ static int process_event_tail(ASS_Track *track, ASS_Event *event,
         NEXT(p, token);
 
         ALIAS(End, Duration)    // temporarily store end timecode in event->Duration
+        ALIAS(Actor, Name)      // both variants are used in files
         PARSE_START
             INTVAL(Layer)
             STYLEVAL(Style)
@@ -579,6 +580,14 @@ static int process_style(ASS_Track *track, char *str)
 
 static bool format_line_compare(const char *fmt1, const char *fmt2)
 {
+#define TOKEN_ALIAS1(token, name, alias) \
+    if (token ## _end - token ## _start == sizeof( #alias ) - 1 &&     \
+            !strncmp(token ## _start, #alias, sizeof( #alias ) - 1)) { \
+        token ## _start = #name;                                       \
+        token ## _end = token ## _start + sizeof( #name ) - 1;         \
+    }
+#define TOKEN_ALIAS(name, alias) TOKEN_ALIAS1(tk1, name, alias) TOKEN_ALIAS1(tk2, name, alias)
+
     while (true) {
         const char *tk1_start, *tk2_start;
         const char *tk1_end, *tk2_end;
@@ -591,12 +600,16 @@ static bool format_line_compare(const char *fmt1, const char *fmt2)
         advance_token_pos(&fmt1, &tk1_start, &tk1_end);
         advance_token_pos(&fmt2, &tk2_start, &tk2_end);
 
+        TOKEN_ALIAS(Name, Actor)
         if ((tk1_end-tk1_start) != (tk2_end-tk2_start))
             return false;
         if (ass_strncasecmp(tk1_start, tk2_start, tk1_end-tk1_start))
             return false;
     }
     return *fmt1 == *fmt2;
+
+#undef TOKEN_ALIAS
+#undef TOKEN_ALIAS1
 }
 
 
@@ -919,6 +932,8 @@ static int process_fonts_line(ASS_Track *track, char *str)
             goto mem_fail;
         track->parser_priv->fontdata_size = new_size;
     }
+    if (!track->parser_priv->fontdata)
+        return 0;
     memcpy(track->parser_priv->fontdata + track->parser_priv->fontdata_used,
            str, len);
     track->parser_priv->fontdata_used += len;
@@ -1225,17 +1240,18 @@ out:
 /**
  * \brief read file contents into newly allocated buffer
  * \param fname file name
+ * \param hint file name origin
  * \param bufsize out: file size
  * \return pointer to file contents. Caller is responsible for its deallocation.
  */
-char *read_file(ASS_Library *library, char *fname, size_t *bufsize)
+char *read_file(ASS_Library *library, const char *fname, FileNameSource hint, size_t *bufsize)
 {
     int res;
     long sz;
     long bytes_read;
     char *buf;
 
-    FILE *fp = fopen(fname, "rb");
+    FILE *fp = ass_open_file(fname, hint);
     if (!fp) {
         ass_msg(library, MSGL_WARN,
                 "ass_read_file(%s): fopen failed", fname);
@@ -1360,7 +1376,7 @@ static char *read_file_recode(ASS_Library *library, char *fname,
     char *buf;
     size_t bufsize;
 
-    buf = read_file(library, fname, &bufsize);
+    buf = read_file(library, fname, FN_EXTERNAL, &bufsize);
     if (!buf)
         return 0;
 #ifdef CONFIG_ICONV
@@ -1416,7 +1432,7 @@ int ass_read_styles(ASS_Track *track, char *fname, char *codepage)
     ParserState old_state;
     size_t sz;
 
-    buf = read_file(track->library, fname, &sz);
+    buf = read_file(track->library, fname, FN_EXTERNAL, &sz);
     if (!buf)
         return 1;
 #ifdef CONFIG_ICONV
@@ -1522,17 +1538,34 @@ fail:
 
 int ass_track_set_feature(ASS_Track *track, ASS_Feature feature, int enable)
 {
+    if (feature >= sizeof(track->parser_priv->feature_flags) * CHAR_BIT || feature < 0)
+        return -1;
+
+    // all supported non-meta features
+    static const uint32_t supported =
+#ifdef USE_FRIBIDI_EX_API
+        FEATURE_MASK(ASS_FEATURE_BIDI_BRACKETS) |
+#endif
+        FEATURE_MASK(ASS_FEATURE_WHOLE_TEXT_LAYOUT) |
+        0;
+    uint32_t requested = 0;
+
     switch (feature) {
     case ASS_FEATURE_INCOMPATIBLE_EXTENSIONS:
-        //-fallthrough
-#ifdef USE_FRIBIDI_EX_API
-    case ASS_FEATURE_BIDI_BRACKETS:
-        track->parser_priv->bidi_brackets = !!enable;
-#endif
-        return 0;
+        requested = supported;
+        break;
     default:
-        return -1;
+        if (!(FEATURE_MASK(feature) & supported))
+            return -1;
+        requested = FEATURE_MASK(feature);
     }
+
+    if (enable)
+        track->parser_priv->feature_flags |= requested;
+    else
+        track->parser_priv->feature_flags &= ~requested;
+
+    return 0;
 }
 
 /**
