@@ -21,6 +21,8 @@
 
 #include <stdarg.h>
 
+#include "ass_filesystem.h"
+
 typedef struct {
     char *name;
     char *data;
@@ -38,6 +40,6 @@ struct ass_library {
     void *msg_callback_data;
 };
 
-char *read_file(struct ass_library *library, char *fname, size_t *bufsize);
+char *read_file(struct ass_library *library, const char *fname, FileNameSource hint, size_t *bufsize);
 
 #endif                          /* LIBASS_LIBRARY_H */
