@@ -51,7 +51,7 @@ static inline int ilog2(uint32_t n)
 }
 
 
-bool rasterizer_init(RasterizerData *rst, int tile_order, int outline_error)
+bool rasterizer_init(const BitmapEngine *engine, RasterizerData *rst, int outline_error)
 {
     rst->outline_error = outline_error;
     rst->linebuf[0] = rst->linebuf[1] = NULL;
@@ -59,7 +59,9 @@ bool rasterizer_init(RasterizerData *rst, int tile_order, int outline_error)
     rst->size[1] = rst->capacity[1] = 0;
     rst->n_first = 0;
 
-    rst->tile = ass_aligned_alloc(32, 1 << (2 * tile_order), false);
+    unsigned align = 1 << engine->align_order;
+    unsigned size = 1 << (2 * engine->tile_order);
+    rst->tile = ass_aligned_alloc(align, size, false);
     return rst->tile;
 }
 
@@ -312,7 +314,7 @@ bool rasterizer_set_outline(RasterizerData *rst,
             return false;
         }
     }
-    assert(start == cur && cur == path->points + path->n_points);
+    assert(start == cur && (!cur || cur == path->points + path->n_points));
 
     for (size_t k = rst->n_first; k < rst->size[0]; k++) {
         struct segment *line = &rst->linebuf[0][k];
@@ -674,8 +676,7 @@ static bool rasterizer_fill_level(const BitmapEngine *engine, RasterizerData *rs
         else
             engine->fill_halfplane(rst->tile, width, line1->a, line1->b, line1->c,
                                    flags1 & FLAG_REVERSE ? -line1->scale : line1->scale);
-        // XXX: better to use max instead of add
-        engine->add_bitmaps(buf, stride, rst->tile, width, height, width);
+        engine->merge_tile(buf, stride, rst->tile);
         rst->size[index] = offs;
         return true;
     }
